from typing import (
    Any,
    Dict,
    Union,
    cast,
)

from eth_utils import (
    keccak,
)
from eth_utils.toolz import (
    pipe,
)
from hexbytes import (
    HexBytes,
)
import rlp
from typing_extensions import Self  # move to typing when >= py311


class HashableRLP(rlp.Serializable):  # type: ignore
    r"""
    An extension of :class:`rlp.Serializable`. In addition to the below
    functions, the class is iterable.

    Use like:

    ::

        class MyRLP(HashableRLP):
            fields = (
                ('name1', rlp.sedes.big_endian_int),
                ('name2', rlp.sedes.binary),
                etc...
            )

        my_obj = MyRLP(name2=b'\xff', name1=1)
        list(my_obj) == [1, b'\xff']
        # note that the iteration order is always in RLP-defined order
    """

    @classmethod
    def from_dict(cls, field_dict: Dict[str, Any]) -> Self:
        r"""
        In addition to the standard initialization of.

        ::

            my_obj = MyRLP(name1=1, name2=b'\xff')

        This method enables initialization with.

        ::

            my_obj = MyRLP.from_dict({'name1': 1, 'name2': b'\xff'})

        In general, the standard initialization is preferred, but
        some approaches might favor this API, like when using
        :meth:`toolz.functoolz.pipe`.

        ::

            return eth_utils.toolz.pipe(
                my_dict,
                normalize,
                validate,
                MyRLP.from_dict,
            )

        :param dict field_dict: the dictionary of values to initialize with
        :returns: the new rlp object
        :rtype: HashableRLP
        """
        return cls(**field_dict)

    @classmethod
    def from_bytes(cls, serialized_bytes: Union[bytes, bytearray]) -> Self:
        """
        Shorthand invocation for :meth:`rlp.decode` using this class.

        :param bytes serialized_bytes: the byte string to decode
        :return: the decoded object
        :rtype: HashableRLP
        """
        decoded = rlp.decode(serialized_bytes, cls)
        return cast(Self, decoded)

    def hash(self) -> HexBytes:
        """
        :returns: the hash of the encoded bytestring
        :rtype: ~hexbytes.main.HexBytes
        """
        return pipe(
            self,
            rlp.encode,
            keccak,
            HexBytes,
        )

    def __iter__(self) -> Any:
        if hasattr(self, "fields"):
            return iter(getattr(self, field) for field, _ in self.fields)
        else:
            return super().__iter__()

    def as_dict(self) -> Dict[str, Any]:
        """
        Convert rlp object to a dict

        :returns: mapping of RLP field names to field values
        :rtype: dict
        """
        try:
            _as_dict = super().as_dict()
            return cast(Dict[str, Any], _as_dict)
        except AttributeError:
            _as_dict = vars(self)
            return cast(Dict[str, Any], _as_dict)



import pytest
from rlp.sedes import (
    big_endian_int,
)

from eth_rlp import (
    HashableRLP,
)


class TwoIntRLP(HashableRLP):
    fields = (
        ("pos1", big_endian_int),
        ("pos2", big_endian_int),
    )


@pytest.mark.parametrize(
    "unserialized, expected_hash_hex",
    (([0, 1], "0x698b0c299d3182774bd859102bea2f205f0a1b3674c8d1d7aee6b17122a2f73a"),),
)
def test_hash(unserialized, expected_hash_hex):
    rlp_obj = TwoIntRLP(*unserialized)
    assert rlp_obj.hash().hex() == expected_hash_hex


@pytest.mark.parametrize(
    "unserialized, expected_hash_hex",
    (
        (
            {"pos1": 0, "pos2": 1},
            "0x698b0c299d3182774bd859102bea2f205f0a1b3674c8d1d7aee6b17122a2f73a",
        ),
    ),
)
def test_create_with_dict(unserialized, expected_hash_hex):
    assert TwoIntRLP.from_dict(unserialized).hash().hex() == expected_hash_hex


@pytest.mark.parametrize(
    "serialized, expected_hash_hex, expected_fields",
    (
        (
            b"\xc2\x80\x01",
            "0x698b0c299d3182774bd859102bea2f205f0a1b3674c8d1d7aee6b17122a2f73a",
            {"pos1": 0, "pos2": 1},
        ),
    ),
)
def test_from_bytes(serialized, expected_hash_hex, expected_fields):
    rlp_obj = TwoIntRLP.from_bytes(serialized)
    assert rlp_obj.hash().hex() == expected_hash_hex
    for name, val in expected_fields.items():
        assert getattr(rlp_obj, name) == val


@pytest.mark.parametrize(
    "unserialized, expected_fields",
    (
        (
            {"pos2": 1, "pos1": 0},
            [0, 1],
        ),
    ),
)
def test_iteration(unserialized, expected_fields):
    rlp_obj = TwoIntRLP(**unserialized)
    for actual, expected in zip(rlp_obj, expected_fields):
        assert actual == expected
    assert list(rlp_obj) == expected_fields


@pytest.mark.parametrize(
    "unserialized, expected_dict",
    (
        (
            {"pos2": 1, "pos1": 0},
            {"pos2": 1, "pos1": 0},
        ),
    ),
)
def test_to_dict(unserialized, expected_dict):
    rlp_obj = TwoIntRLP(**unserialized)
    assert rlp_obj.as_dict() == expected_dict
